// PV112 2017, lesson 1
// Final version; it is possible that some groups didn't complete everything

#include <memory>
#include <iostream>
#include <sstream>
#include <fstream>

#define GLEW_STATIC
#include "GL/glew.h"
#if defined(_WIN32)
#define NOMINMAX // Make Windows.h not define 'min' and 'max' macros
#include <GL/wglew.h> // Include on Windows
#pragma comment(lib, "glew32s.lib")
#else
#include <GL/glxew.h> // Include on Linux and Mac
#define APIENTRY
#endif

#include "GL/freeglut.h"

using namespace std;

const int vertices_count = 9 + 18;
const float my_positions[vertices_count * 2] = {
    // Initial Triangle:
    //0.0f,  0.5f,
    //-0.5f, -0.5f,
    //0.5f, -0.5f,

    // House:
     0.0f,  0.6f,
    -0.4f,  0.2f,
     0.4f,  0.2f,

    0.4f,  0.2f,
    -0.4f,  0.2f,
    -0.4f, -0.6f,

     0.4f,  0.2f,
    -0.4f, -0.6f,
     0.4f, -0.6f,
     // Sun (18 vertices)
    -0.80f, 0.80f,        // Center
    -0.80f, 0.95f,        // v1
    -0.77f, 0.87f,        // v2
    -0.70f, 0.90f,        // v3
    -0.73f, 0.83f,        // v4
    -0.65f, 0.80f,        // v5
    -0.73f, 0.77f,        // v6
    -0.70f, 0.70f,        // v7
    -0.77f, 0.73f,        // v8
    -0.80f, 0.65f,        // v9
    -0.83f, 0.73f,        // v10
    -0.90f, 0.70f,        // v11
    -0.87f, 0.77f,        // v12
    -0.95f, 0.80f,        // v13
    -0.87f, 0.83f,        // v14
    -0.90f, 0.90f,        // v15
    -0.83f, 0.87f,        // v16
    -0.80f, 0.95f,        // v1 again -> closing the triangle fan
};

const unsigned char my_colors[vertices_count * 3] = {
    // Initial triangle
    //0, 255, 255,
    //128, 255, 0,
    //255, 255, 255

    // House:
    255, 0, 0,
    255, 0, 0,
    255, 0, 0,

    255, 255, 0,
    255, 255, 0,
    255, 255, 0,

    255, 255, 0,
    255, 255, 0,
    255, 255, 0,

    // Sun:
    255, 255, 127,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
    64, 64, 0,
};

// Not used anymore, we switched back to triangles
/*
const unsigned indices_count = 9;
const unsigned my_indices [indices_count] {
    0, 1, 2,
    1, 2, 3,
    4, 3, 2,
};
*/

int win_width = 480;
int win_height = 480;

// Identifiers of buffers, VAO and GLSL program
GLuint positions_vbo;
GLuint colors_vbo;
GLuint indices_vbo;
GLuint my_geometry_vao;
GLuint program;

// Locations of shader variables
GLint position_location;
GLint color_location;
GLint aspect_location;

// Callback function to be called when a key is pressed
void key_pressed(unsigned char key, int mouseX, int mouseY)
{
    switch (key)
    {
    case 27: // Escape
        exit(0);
        break;
    case 'l':
        // Wireframe mode
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glutPostRedisplay();
        break;
    case 'f':
        // Fill mode
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glutPostRedisplay();
        break;
    case 't':
        // Full screen
        glutFullScreenToggle();
        break;
    }
}

// This function loads a file and returns its content as std::string
string load_file_to_string(const char *file_name)
{
    ifstream file(file_name);
    stringstream ss;
    ss << file.rdbuf();
    return ss.str();
}

void init()
{
    // Background color
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    // Load and compile vertex shader
    int compile_status;
    const char *source;

    string vs_source = load_file_to_string("vertex.glsl");
    GLuint vs_shader = glCreateShader(GL_VERTEX_SHADER);
    source = vs_source.c_str();
    glShaderSource(vs_shader, 1, &source, nullptr);
    glCompileShader(vs_shader);
    // Check for errors
    glGetShaderiv(vs_shader, GL_COMPILE_STATUS, &compile_status);
    if (GL_FALSE == compile_status)
    {
        int log_len = 0;
        glGetShaderiv(vs_shader, GL_INFO_LOG_LENGTH, &log_len);
        unique_ptr<char []> log(new char[log_len]);
        glGetShaderInfoLog(vs_shader, log_len, nullptr, log.get());
        cout << "Failed to compile vertex shader" << endl;
        cout << log.get() << endl;

        getchar();        // Wait for keypress
        exit(1);
    }

    string fs_source = load_file_to_string("fragment.glsl");
    GLuint fs_shader = glCreateShader(GL_FRAGMENT_SHADER);
    source = fs_source.c_str();
    glShaderSource(fs_shader, 1, &source, nullptr);
    glCompileShader(fs_shader);
    // Check for errors
    glGetShaderiv(fs_shader, GL_COMPILE_STATUS, &compile_status);
    if (GL_FALSE == compile_status)
    {
        int log_len = 0;
        glGetShaderiv(fs_shader, GL_INFO_LOG_LENGTH, &log_len);
        unique_ptr<char []> log(new char[log_len]);
        glGetShaderInfoLog(fs_shader, log_len, nullptr, log.get());
        cout << "Failed to compile fragment shader" << endl;
        cout << log.get() << endl;

        getchar();        // Wait for keypress
        exit(1);
    }

    // Create program and link it with shaders
    program = glCreateProgram();
    glAttachShader(program, vs_shader);
    glAttachShader(program, fs_shader);
    glLinkProgram(program);
    // Check for errors
    glGetProgramiv(program, GL_LINK_STATUS, &compile_status);
    if (GL_FALSE == compile_status)
    {
        int log_len = 0;
        glGetProgramiv(program, GL_INFO_LOG_LENGTH, &log_len);
        unique_ptr<char []> log(new char[log_len]);
        glGetProgramInfoLog(program, log_len, nullptr, log.get());
        cout << "Failed to link program" << endl;
        cout << log.get() << endl;

        getchar();        // Wait for keypress
        exit(1);
    }

    // Get attribute (attribute = "in" variable in shader) and uniform locations
    position_location = glGetAttribLocation(program, "position");
    color_location = glGetAttribLocation(program, "color");
    aspect_location = glGetUniformLocation(program, "aspect");

    // Create buffers with geometry and color
    glGenBuffers(1, &positions_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, positions_vbo);
    glBufferData(GL_ARRAY_BUFFER, vertices_count * sizeof(float) * 2, my_positions, GL_STATIC_DRAW);

    glGenBuffers(1, &colors_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, colors_vbo);
    glBufferData(GL_ARRAY_BUFFER, vertices_count * sizeof(unsigned char) * 3, my_colors, GL_STATIC_DRAW);

    glBindBuffer(GL_ARRAY_BUFFER, 0);

    /*
    // Create a buffer for indices
    glGenBuffers(1, &indices_vbo);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indices_vbo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices_count * sizeof(unsigned), my_indices, GL_STATIC_DRAW);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    */

    // Create a vertex array object for the geometry
    glGenVertexArrays(1, &my_geometry_vao);

    // Set the parameters of the geometry
    glBindVertexArray(my_geometry_vao);

    // Bind geometry buffer to the VAO
    glBindBuffer(GL_ARRAY_BUFFER, positions_vbo);
    // Enable the 'position' attribute in shader, linking it with geometry VBO
    glEnableVertexAttribArray(position_location);
    // Define the format of the buffered data (two floats, normalized)
    glVertexAttribPointer(position_location, 2, GL_FLOAT, GL_FALSE, 0, 0);

    // The same for color buffer
    glBindBuffer(GL_ARRAY_BUFFER, colors_vbo);
    glEnableVertexAttribArray(color_location);
    // Colors are 0..255, but shader wants 0.0f..1.0f -> normalisation needed)
    glVertexAttribPointer(color_location, 3, GL_UNSIGNED_BYTE, GL_TRUE, 0, 0);

    // Bind index buffer to the VAO
    //glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indices_vbo);

    glBindVertexArray(0);
}

// Callback function to be called when we want to redraw the scene
void render()
{
    // Fill the framebuffer with background color
    glClear(GL_COLOR_BUFFER_BIT);

    glUseProgram(program);

    // Set the uniform 'aspect' to current window aspect ratio
    glUniform1f(aspect_location, float(win_width) / float(win_height));

    // We want to draw what we defined in the VAO
    glBindVertexArray(my_geometry_vao);
    // Draw house (first 9 vertices -> 3 triangles)
    glDrawArrays(GL_TRIANGLES, 0, 9);
    // Draw sun (18 vertices -> 16 triangles in a fan)
    glDrawArrays(GL_TRIANGLE_FAN, 9, 18);

    // Draw using indices
    //glDrawElements(GL_TRIANGLES, 9, GL_UNSIGNED_INT, 0);

    glBindVertexArray(0);
    glUseProgram(0);

    // Swap front and back buffer -> show our wonderful scenery in the window
    glutSwapBuffers();
}

// Callback function to be called when the window is resized
void reshape(int width, int height)
{
    // Update current size (needed to calculate aspect ratio)
    win_width = width;
    win_height = height;

    // Make GL use the whole window
    glViewport(0, 0, win_width, win_height);
}

// Callback function to be called when we make an error in OpenGL
void APIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
    switch (type)
    {
    case GL_DEBUG_TYPE_ERROR:
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
        cout << message << endl;
        return;
    default:
        return;
    }
}

int main(int argc, char ** argv)
{
    // Initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS); // Return from the main loop

    // Set OpenGL Context parameters
    glutInitContextVersion(3, 3);              // Use OpenGL 3.3
    glutInitContextProfile(GLUT_CORE_PROFILE); // Use OpenGL core profile
    glutInitContextFlags(GLUT_DEBUG);          // Use OpenGL debug context

    // Initialize window
    glutInitWindowSize(win_width, win_height);
    glutCreateWindow("PV112 - cv1");

    // Initialize GLEW
    glewExperimental = GL_TRUE;
    glewInit();

    // Setup callback that will inform us when we make an error.
    // glDebugMessageCallbackARB is sometimes missed by glew, due to a bug in it.
    PFNGLDEBUGMESSAGECALLBACKARBPROC myglDebugMessageCallbackARB;
#if defined(_WIN32) // Windows
    myglDebugMessageCallbackARB =
      (PFNGLDEBUGMESSAGECALLBACKARBPROC)wglGetProcAddress("glDebugMessageCallbackARB");
#elif defined(__APPLE__) // MacOS (not tested)
    myglDebugMessageCallbackARB = glDebugMessageCallbackARB;
#else // Linux
    myglDebugMessageCallbackARB =
      (PFNGLDEBUGMESSAGECALLBACKARBPROC)glXGetProcAddress((unsigned char *)"glDebugMessageCallbackARB");
#endif
    if (myglDebugMessageCallbackARB)
    {
        glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
        myglDebugMessageCallbackARB(simple_debug_callback, nullptr);
    }

    // Initialise our GL stuff (shaders, buffers, ...)
    init();

    // Register callbacks
    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key_pressed);

    // Main loop (returns when the window is closed)
    glutMainLoop();

    return 0;
}

// If you use this structure, you can store position AND color in one buffer.
// That's what you use 'stride' and 'pointer' in glVertexAttribPointer for.

/*
struct my_vertex_struct
{
    float position[2];
    unsigned char color[3];
    unsigned char _padding[1];
};

...
glEnableVertexAttribArray(position_location);
glVertexAttribPointer(position_location, 2, GL_FLOAT, GL_FALSE, sizeof(my_vertex_struct), 0);
glEnableVertexAttribArray(color_location);
glVertexAttribPointer(color_location, 3, GL_UNSIGNED_BYTE, GL_TRUE, sizeof(my_vertex_struct), (const void *)(sizeof(float)*2));
...

const my_vertex_struct my_whole_geometry[vertices_count] = {
     0.0f,  0.6f,        255, 0, 0,      0,
    -0.4f,  0.2f,        255, 0, 0,      0,
     0.4f,  0.2f,        255, 0, 0,      0,
     0.4f,  0.2f,        255, 255, 0,    0,
    -0.4f,  0.2f,        255, 255, 0,    0,
    -0.4f, -0.6f,        255, 255, 0,    0,
     0.4f, -0.6f,        255, 255, 0,    0,
    -0.80f, 0.80f,       255, 255, 127,  0,
    -0.80f, 0.95f,       0, 0, 0,        0,
    -0.77f, 0.87f,       0, 0, 0,        0,
    -0.70f, 0.90f,       0, 0, 0,        0,
    -0.73f, 0.83f,       0, 0, 0,        0,
    -0.65f, 0.80f,       0, 0, 0,        0,
    -0.73f, 0.77f,       0, 0, 0,        0,
    -0.70f, 0.70f,       0, 0, 0,        0,
    -0.77f, 0.73f,       0, 0, 0,        0,
    -0.80f, 0.65f,       0, 0, 0,        0,
    -0.83f, 0.73f,       0, 0, 0,        0,
    -0.90f, 0.70f,       0, 0, 0,        0,
    -0.87f, 0.77f,       0, 0, 0,        0,
    -0.95f, 0.80f,       0, 0, 0,        0,
    -0.87f, 0.83f,       0, 0, 0,        0,
    -0.90f, 0.90f,       0, 0, 0,        0,
    -0.83f, 0.87f,       0, 0, 0,        0,
    -0.80f, 0.95f,       0, 0, 0,        0,
};
*/
